@@ -34,9 +34,10 @@
 
 #include "CopcReader.hpp"
 
+#include <algorithm>
+#include <atomic>
 #include <functional>
 #include <limits>
-#include <algorithm>
 
 #include <nlohmann/json.hpp>
 
@@ -45,6 +46,7 @@
 #include <pdal/Polygon.hpp>
 #include <pdal/Scaling.hpp>
 #include <pdal/SrsBounds.hpp>
+#include <pdal/private/OGRSpec.hpp>
 #include <pdal/util/Charbuf.hpp>
 #include <pdal/util/ThreadPool.hpp>
 #include <pdal/private/gdal/GDALUtils.hpp>
@@ -222,9 +224,7 @@ public:
     bool fixNames;
     bool doVlrs;
 
-    NL::json query;
-    NL::json headers;
-    NL::json ogr;
+    OGRSpec ogr;
 
     int keepAliveChunkCount = 10;
     SrsOrderSpec srsVlrOrder;
@@ -234,14 +234,14 @@ public:
 struct CopcReader::Private
 {
 public:
+    std::mutex mutex;
+    std::queue<copc::Tile> contents;
     std::unique_ptr<ThreadPool> pool;
     std::unique_ptr<copc::Tile> currentTile;
 
     std::unique_ptr<connector::Connector> connector;
-    std::queue<copc::Tile> contents;
     copc::Hierarchy hierarchy;
     las::LoaderDriver loader;
-    std::mutex mutex;
     std::condition_variable contentsCv;
     std::condition_variable consumedCv;
     std::vector<PolyXform> polys;
@@ -255,7 +255,7 @@ public:
     las::Header header;
     copc::Info copc_info;
     point_count_t hierarchyPointCount;
-    bool done;
+    std::atomic<bool> done;
     SrsTransform llToBcbfTransform;
 };
 
@@ -264,7 +264,12 @@ CopcReader::CopcReader() : m_args(new CopcReader::Args), m_p(new CopcReader::Pri
 
 
 CopcReader::~CopcReader()
-{}
+{
+    // We join the pool rather than let the dtor do it because the workers use m_p
+    // which will get 0'ed in the dtor before access through m_p is complete in
+    // the workers.
+    done();
+}
 
 
 std::string CopcReader::getName() const
@@ -293,8 +298,6 @@ void CopcReader::addArgs(ProgramArgs& args)
     args.add("resolution", "Resolution limit", m_args->resolution);
     args.add("polygon", "Bounding polygon(s) to crop requests",
         m_args->polys).setErrorText("Invalid polygon specification. Must be valid GeoJSON/WKT");
-    args.add("header", "Header fields to forward with HTTP requests", m_args->headers);
-    args.add("query", "Query parameters to forward with HTTP requests", m_args->query);
     args.add("ogr", "OGR filter geometries", m_args->ogr);
     args.add("fix_dims", "Make invalid dimension names valid by changing invalid "
         "characters to '_'", m_args->fixNames, true);
@@ -307,42 +310,15 @@ void CopcReader::addArgs(ProgramArgs& args)
 }
 
 
-void CopcReader::setForwards(StringMap& headers, StringMap& query)
-{
-    try
-    {
-        if (!m_args->headers.is_null())
-            headers = m_args->headers.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'headers': ") + err.what());
-    }
-
-    try
-    {
-        if (!m_args->query.is_null())
-            query = m_args->query.get<StringMap>();
-    }
-    catch (const std::exception& err)
-    {
-        throwError(std::string("Error parsing 'query': ") + err.what());
-    }
-}
-
-
 void CopcReader::initialize(PointTableRef table)
 {
-    const std::size_t threads(m_args->threads);
-    if (threads > 100)
+    if (m_args->threads > 100)
         log()->get(LogLevel::Warning) << "Using a large thread count: " <<
-            threads << " threads" << std::endl;
-    m_p->pool.reset(new ThreadPool(threads));
+            m_args->threads << " threads" << std::endl;
+    // Make sure we allow at least as many chunks as we have threads.
+    m_args->keepAliveChunkCount = (std::max)(m_args->threads, (size_t)m_args->keepAliveChunkCount);
 
-    StringMap headers;
-    StringMap query;
-    setForwards(headers, query);
-    m_p->connector.reset(new connector::Connector(m_filename, headers, query));
+    m_p->connector.reset(new connector::Connector(m_filespec));
 
     MetadataNode forward = table.privateMetadata("lasforward");
     MetadataNode m = getMetadata();
@@ -406,10 +382,13 @@ void CopcReader::initialize(PointTableRef table)
     m_p->depthEnd = m_args->resolution ?
         (std::max)(1, (int)ceil(log2(m_p->copc_info.spacing / m_args->resolution)) + 1) :
         0;
-    
+
     if (m_args->resolution)
-        log()->get(LogLevel::Debug) << "Maximum depth: " << m_p->depthEnd << 
-            std::endl;
+        log()->get(LogLevel::Debug) << "Maximum depth: " << m_p->depthEnd << std::endl;
+
+    // Initialize our threadpool
+    m_p->pool.reset(new ThreadPool(m_args->threads));
+
 }
 
 
@@ -490,7 +469,7 @@ las::VlrList CopcReader::fetchSrsVlrs(const las::VlrCatalog& catalog)
     fetchVlr(las::TransformUserId, las::LASFWkt2recordId);
     fetchVlr(las::PdalUserId, las::PdalProjJsonRecordId);
     fetchVlr(las::TransformUserId, las::WktRecordId);
-    
+
     // User told us to ditch them
     if (m_args->nosrs)
         vlrs.clear();
@@ -568,13 +547,8 @@ void CopcReader::createSpatialFilters()
         }
     }
 
-    // Read polygons from OGR and add to the polygon list.
-    if (!m_args->ogr.is_null())
-    {
-        auto& plist = m_args->polys;
-        std::vector<Polygon> ogrPolys = gdal::getPolygons(m_args->ogr);
-        plist.insert(plist.end(), ogrPolys.begin(), ogrPolys.end());
-    }
+    std::vector<Polygon> ogrPolys = m_args->ogr.getPolygons();
+    m_args->polys.insert(m_args->polys.end(), ogrPolys.begin(), ogrPolys.end());
 
     // Create transform from the point source SRS to the poly SRS.
     for (Polygon& poly : m_args->polys)
@@ -688,12 +662,11 @@ void CopcReader::ready(PointTableRef table)
         totalPoints += entry.m_pointCount;
 
     if (totalPoints > 1e8)
-        log()->get(LogLevel::Warning) << totalPoints << " will be downloaded" << std::endl;
+        log()->get(LogLevel::Warning) << totalPoints << " points will be downloaded" << std::endl;
 
     m_p->tileCount = m_p->hierarchy.size();
     log()->get(LogLevel::Debug) << m_p->tileCount << " overlapping nodes" << std::endl;
 
-    m_p->pool.reset(new ThreadPool(m_p->pool->numThreads()));
     m_p->done = false;
     for (const copc::Entry& entry : m_p->hierarchy)
         load(entry);
@@ -832,10 +805,9 @@ void CopcReader::load(const copc::Entry& entry)
 
             // Put the tile on the output queue.
             std::unique_lock<std::mutex> l(m_p->mutex);
-            if (m_p->done)
-                return;
-            while (m_p->contents.size() >= (std::max)((size_t)m_args->keepAliveChunkCount, m_p->pool->numThreads()))
-                m_p->consumedCv.wait(l);
+            m_p->consumedCv.wait(l, [this] {
+                return (m_p->done ||
+                    m_p->contents.size() < (size_t)m_args->keepAliveChunkCount);});
             m_p->contents.push(std::move(tile));
             l.unlock();
             m_p->contentsCv.notify_one();
@@ -1004,13 +976,18 @@ top:
 
 void CopcReader::done(PointTableRef)
 {
+    done();
+}
+
+void CopcReader::done()
+{
+    if (m_p->pool)
     {
-        std::unique_lock<std::mutex> l(m_p->mutex);
         m_p->done = true;
+        m_p->consumedCv.notify_all();
+        m_p->pool->stop();
+        m_p->connector.reset();
     }
-    m_p->consumedCv.notify_all();
-    m_p->pool->stop();
-    m_p->connector.reset();
 }
 
 } // namespace pdal
